import abc
from functools import partial
from functools import wraps
from typing import Annotated
from typing import Any
from typing import Callable
from typing import Final
from typing import Iterable
from typing import TypeVar
from typing import get_args

from .type_hints import extract_annotated
from .type_hints import get_resolvable_type_hints

_abstract_marker: Final = object()
_O = TypeVar("_O")
Abstract = Annotated[_O, _abstract_marker]


def get_abstract_attributes(cls: type) -> Iterable[tuple[str, type]]:
    for var, hint in get_resolvable_type_hints(cls).items():
        annotated = extract_annotated(hint)
        if not annotated:
            continue
        # Checking for both the abstract marker and the type alias itself allows both
        # a concise way using e.g. `var: Abstract[int]` as well as a verbose way that
        # allows combining the qualifier with other annotated types and qualifiers, e.g.
        # `var: Annotated[int, Abstract, Other]`.
        if {_abstract_marker, Abstract} & set(get_args(annotated)):
            yield var, annotated


C = TypeVar("C")


def abstractattrs(cls: C) -> C:
    attributes = get_abstract_attributes(cls)  # type: ignore[arg-type]

    # Save metadata about the abstract attributes on the decorated class. By merging
    # with any already defined attributes we support multiple decorations and
    # inheritance.
    cls.__abstract_attributes__ = frozenset(  # type: ignore[attr-defined]
        {var for var, _ in attributes} | getattr(cls, "__abstract_attributes__", set())
    )

    # Add an __init_subclass__ method to the base class. If it already defines one, we
    # wrap it inside a function that will first check the subclasses' class attributes
    # and then call the already existing method.
    cls.__init_subclass__ = (  # type: ignore[assignment]
        classmethod(  # type: ignore[assignment]
            wraps(cls.__init_subclass__)(
                partial(
                    _init_subclass,
                    existing_init_subclass=(
                        cls.__init_subclass__.__func__  # type: ignore[attr-defined]
                    ),
                )
            )
        )
        if "__init_subclass__" in cls.__dict__
        else classmethod(_init_subclass)
    )

    return cls


class UndefinedAbstractAttribute(TypeError):
    ...


def check_abstract_class_attributes(cls: type) -> None:
    """Check that a class defines inherited abstract attributes."""
    if abc.ABC in cls.__bases__:
        return

    for attr in getattr(cls, "__abstract_attributes__", ()):
        if not hasattr(cls, attr):
            raise UndefinedAbstractAttribute(
                f"Concrete class {cls.__qualname__!r} must define abstract "
                f"class attribute {attr!r}."
            )


def _init_subclass(
    cls: type,
    existing_init_subclass: Callable[..., None] | None = None,
    *args: Any,
    **kwargs: Any,
) -> None:
    """
    __init_subclass__ implementation added to classes decorated with @abstractattrs.
    """
    super(cls).__init_subclass__()  # type: ignore[misc]
    check_abstract_class_attributes(cls)
    if existing_init_subclass is not None:
        existing_init_subclass(cls, *args, **kwargs)



from typing import Annotated
from typing import ClassVar
from typing import Final
from typing import ForwardRef
from typing import get_args
from typing import get_origin
from typing import get_type_hints

max_iterations: Final = 10_000


class MaxIterations(RuntimeError):
    ...


def get_name_error_name(error: NameError) -> str:
    return str(error).split("'", 2)[1]


def get_resolvable_type_hints(cls: type) -> dict[str, type]:
    """
    Repeatedly attempt to resolve the type hints of the given class, handling the
    NameErrors produced by unresolvable names by inserting sentinel typing.ForwardRef
    objects.
    """
    local_ns = {cls.__name__: cls}

    for _ in range(max_iterations):
        try:
            return get_type_hints(cls, include_extras=True, localns=local_ns)
        except NameError as exception:
            name = get_name_error_name(exception)
            local_ns[name] = ForwardRef(name)  # type: ignore[assignment]
    else:  # pragma: no cover
        raise MaxIterations(
            f"Exceeded {max_iterations} iterations trying to resolve type hints of "
            f"{cls.__module__}.{cls.__qualname__}."
        )


def extract_annotated(hint: type) -> type | None:
    # Unwrap ClassVar[T] -> T.
    if get_origin(hint) is ClassVar:
        try:
            (hint,) = get_args(hint)
        # Wrong-argument ClassVars are very hard to produce at runtime, so pragma is
        # reasonable here.
        except ValueError:  # pragma: no cover
            return None

    if not get_origin(hint) is Annotated:
        return None

    return hint



from .abcattrs import Abstract
from .abcattrs import UndefinedAbstractAttribute
from .abcattrs import abstractattrs
from .abcattrs import check_abstract_class_attributes

__all__ = (
    "Abstract",
    "UndefinedAbstractAttribute",
    "abstractattrs",
    "check_abstract_class_attributes",
)
__version__ = "0.3.2"
